#ifndef PERSISTENTSTORAGE_H
#define PERSISTENTSTORAGE_H

#include "SSTypes.h"
#include "BufferPool.h"
#include "MsgSS.h" //For MAX_MSGSS_PAYLOAD_SIZE

#define EE_STREAM_START_PAGE  (EEPROM_LOGGER_APPEND_START + 1)
#define EE_STREAM_END_PAGE    (EEPROM_LOGGER_APPEND_END)
#define EE_USABLE_SPACE       (EE_STREAM_END_PAGE - EE_STREAM_START_PAGE)
#define EE_HEADER_PAGE        (EEPROM_LOGGER_APPEND_START)

#define atomic_ph if(1)

sstor_header sh;              // Global storage header structure

uint8_t ee_state = EE_NOT_OPENED;
uint8_t reads_enqueued = 0;   // Reads to be done
bool ee_write_mode;           // write or read mode

int8_t ee_buffer[TOS_EEPROM_LINE_SIZE * 2];
uint8_t* ee_bufferA = (int8_t*) &(ee_buffer[0]);
uint8_t* ee_bufferB = (int8_t*) &(ee_buffer[TOS_EEPROM_LINE_SIZE]);

inline uint8_t checksum(sstor_header* h) {
  uint8_t i,s,c;

  c = h->checksum;
  h->checksum = 0;
  s = 0;
  for (i=0; i < sizeof(sstor_header); i++)
    s += (uint8_t) *(((uint8_t*) h) + i);
  h->checksum = c;
  dbg(DBG_USR3, "checksum: %d\n", s);
  return s;
}

/***********
  PS Macros
************/

inline result_t ps_format() {
  uint8_t i;
  
  atomic_ph {
		if (ee_state != EE_NOT_OPENED) {
			dbg(DBG_USR3, "ps_format: ERROR! cannot format, eeprom is in use\n");
			return FAIL;
		}	
		ee_state = EE_WRITE_EMPTY_HEADER;
  }
  
  if (call EEPROMWrite.startWrite() == FAIL) {
    dbg(DBG_USR3, "ps_format: ERROR! Cannot write to eeprom (startWrite fails)\n");
    ee_state = EE_NOT_OPENED;
    return FAIL;  
  }
  
  sh.node = TOS_LOCAL_ADDRESS;
  sh.record_size = 0;
  sh.rpos = 0;    //EE Page alignment
  sh.wpos = 0;    //EE Page alignment
  sh.stored_records = 0;
  for (i=0; i < sizeof(madwise_string); i++)
    sh.name[i] = madwise_string[i];
  sh.checksum = checksum(&sh);
    
  if (call EEPROMWrite.write(EE_HEADER_PAGE, (uint8_t*) &sh) == FAIL) {
    dbg(DBG_USR3, "ps_format: ERROR! Cannot write to eeprom (write fails)\n");
    ee_state = EE_NOT_OPENED;
    return FAIL;
  }
  
  dbg(DBG_USR3, "ps_format: Write requested for header (%d bytes)\n", sizeof(sstor_header));
  return SUCCESS;
}


inline result_t ps_open(bool wmode) {

  if (ee_state == EE_READY)
    if (ee_write_mode == wmode) {
      dbg(DBG_USR3, "ps_open: stream ready opened in this mode\n");
      return SUCCESS;
    } else {
      dbg(DBG_USR3, "ps_open: ERROR: Stream already opened in a different mode\n");
      return FAIL;
    }

  atomic_ph {
		if (ee_state != EE_NOT_OPENED) {
      dbg(DBG_USR3, "ps_open: ERROR: Cannot open, eeprom is in use\n");
	    return FAIL;
		}
		ee_state = EE_READ_HEADER;
  }

  ee_write_mode = wmode;

  if (wmode) {
    dbg(DBG_USR3, "ps_open: stream opened in write mode\n");
  } else {
    dbg(DBG_USR3, "ps_open: stream opened in read mode\n");
  }
  
  if (call EEPROMRead.read(EE_HEADER_PAGE, (uint8_t*) &sh) == FAIL) {
    dbg(DBG_USR3, "ps_open: ERROR! cannot read header\n");
    ee_state = EE_NOT_OPENED;
    return FAIL;
  }
  
  dbg(DBG_USR3, "ps_open: Read requested for header\n");
  return SUCCESS;
}


inline result_t ps_close() {
  
  if (ee_state == EE_NOT_OPENED) {
    dbg(DBG_USR3, "ps_close: ERROR! stream not opened\n");
    return FAIL;
  }

  atomic_ph {
		if (ee_state != EE_READY) {
			dbg(DBG_USR3, "ps_close: ERROR! cannot close, eeprom is in use\n");
			return FAIL;
		}	
		ee_state = EE_WRITE_HEADER;
  }

  sh.checksum = checksum(&sh);

  if (call EEPROMWrite.startWrite() == FAIL) {
    dbg(DBG_USR3, "ps_close: ERROR! Cannot write header to eeprom (startWrite fails)\n");
    ee_state = EE_READY;
    return FAIL;  
  }
  
  if (call EEPROMWrite.write(EE_HEADER_PAGE, (uint8_t*) &sh) == FAIL) {
    dbg(DBG_USR3, "ps_close: ERROR! Cannot write header to eeprom (write fails)\n");
    ee_state = EE_READY;
    return FAIL;
  }

  dbg(DBG_USR3, "ps_close: Write requested for header (%d bytes): node=%d record_size=%d rpos=%d wpos=%d stored_records=%d checksum=%d\n", sizeof(sstor_header), sh.node, sh.record_size, sh.rpos, sh.wpos, sh.stored_records, sh.checksum);
  return SUCCESS;

}

inline result_t ps_read() {

  if (ee_state == EE_NOT_OPENED) {
    dbg(DBG_USR3, "ps_read: ERROR! stream not opened\n");
    return FAIL;
  }

  if (ee_write_mode) {
    dbg(DBG_USR3, "ps_read: ERROR! operation not permitted: stream opened in write mode\n");
    return FAIL;
  }

  if (sh.record_size == 0) {
    dbg(DBG_USR3, "ps_read: ERROR! record_size not yet defined.\n");
    return FAIL;
  }

  if (sh.stored_records == 0) {
    dbg(DBG_USR3, "ps_read: ERROR! eeprom is empty\n");
    return FAIL;
  }

  atomic_ph {
		// Se la eeprom è occupata metto la read in coda
		if (ee_state != EE_READY) {
			dbg(DBG_USR3, "ps_read: eeprom busy, read request enqueued\n");
			reads_enqueued++;
			return SUCCESS;
		}
		ee_state = EE_READ_SINGLE;
  }

  // Record is spread across 2 EE pages?
  if (sh.record_size > TOS_EEPROM_LINE_SIZE) {
    dbg(DBG_USR3, "ps_read: record_size %d greater than eeprom page size %d\n", sh.record_size, TOS_EEPROM_LINE_SIZE);
    ee_state = EE_READ_FIRST; 
  }

  if (call EEPROMRead.read(EE_STREAM_START_PAGE + sh.rpos, ee_bufferA) == FAIL) {
    dbg(DBG_USR3, "ps_read: ERROR! read from eeprom failed!\n");
    ee_state = EE_READY;
    return FAIL;
  }

  dbg(DBG_USR3, "ps_read: read requested\n");
  
  return SUCCESS;
}


inline result_t ps_write(int8_t* data, uint8_t nbytes) {
  uint8_t i;

  if (ee_state == EE_NOT_OPENED) {
    dbg(DBG_USR3, "ps_write: ERROR! stream not opened\n");
    return FAIL;
  }

  if (!ee_write_mode) {
    dbg(DBG_USR3, "ps_write: ERROR! operation not permitted: stream opened in read mode\n");
    return FAIL;
  }
  
  atomic_ph {
		if (ee_state != EE_READY) {
      dbg(DBG_USR3, "ps_write: ERROR! cannot write, eeprom busy (state=%d)\n", ee_state);
			return FAIL;
		}
		ee_state = EE_WRITE_SINGLE;
  }

  if (sh.record_size == 0) {
    if (nbytes > TOS_EEPROM_LINE_SIZE * 2) {
      dbg(DBG_USR3, "ps_write: ERROR! record_size exceed maximum size (maximum expected %d, found %d)\n", TOS_EEPROM_LINE_SIZE*2, nbytes);
      ee_state = EE_READY;
      return FAIL;
    }
    sh.record_size = nbytes;
    dbg(DBG_USR3, "ps_write: record_size defined: %d\n", sh.record_size);
    
  } else {

    if (sh.record_size != nbytes) {
      dbg(DBG_USR3, "ps_write: ERROR! record_size does not match with nbytes (expected %d, found %d)\n", sh.record_size, nbytes);
      ee_state = EE_READY;
      return FAIL;
    }
  }

  //Copy data onto the buffer
  for (i=0; i < sh.record_size; i++)
    ee_buffer[i] = data[i];

  // Record is spread across 2 EE pages?
  if (sh.record_size > TOS_EEPROM_LINE_SIZE) {
    ee_state = EE_WRITE_FIRST;
  }
  
  if (call EEPROMWrite.startWrite() == FAIL) {
    dbg(DBG_USR3, "ps_close: ERROR! Cannot write to eeprom (startWrite fails)\n");
    ee_state = EE_READY;
    return FAIL;  
  }
  
  if (call EEPROMWrite.write(EE_STREAM_START_PAGE + sh.wpos, ee_bufferA) == FAIL) {
    dbg(DBG_USR3, "ps_close: ERROR! Cannot write to eeprom (write fails)\n");
    ee_state = EE_READY;
    return FAIL;
  }
  
  dbg(DBG_USR3, "ps_write: write to eeprom requested (ee_state=%d, record_size=%d, wpos=%d)\n", ee_state, sh.record_size, sh.wpos);

  return SUCCESS;
}


/***********
  EEprom Events
************/

event result_t EEPROMRead.readDone(uint8_t *buffer, result_t success) {
  
  stream_type stream_desc;

  stream_desc.table_id = PSDT_ID;
  stream_desc.no = 0;
    
	switch (ee_state) {
		case EE_READY:
			dbg(DBG_USR3, "EEPROMRead.readDone: ERROR! Inconsistent state! readDone received but no read in progress\n");
		break;

		//Header requested
		case EE_READ_HEADER:
			if (success == SUCCESS) {
				if (checksum(&sh) == sh.checksum) {
          dbg(DBG_USR3, "EEPROMRead.readDone: header fetched: record_size=%d rpos=%d wpos=%d stored_records=%d\n", sh.record_size, sh.rpos, sh.wpos, sh.stored_records);
          ee_state = EE_READY;
          
				} else {	
					dbg(DBG_USR3, "EEPROMRead.readDone: EEprom header corrupted! Format required.\n");
					ee_state = EE_NOT_OPENED;
				}
				
			} else {
				dbg(DBG_USR3, "EEPROMRead.readDone: ERROR: cannot fetch eeprom header\n");
				ee_state = EE_NOT_OPENED;
			}
		break;

		//This page contains the entire record
		case EE_READ_SINGLE:
			if (success == SUCCESS) {
			  dbg(DBG_USR3, "EEPROMRead.readDone: buffer content (first 4 bytes) %d %d %d %d\n", ee_buffer[0], ee_buffer[1], ee_buffer[2], ee_buffer[3]);
				signal SSToApp.readDone(stream_desc, ee_buffer, sh.record_size, SUCCESS);
				call Leds.greenToggle();
				sh.rpos++;
				sh.stored_records--;
				dbg(DBG_USR3, "EEPROMRead.readDone: signaling SSToApp.readDone (table_id=%d no=%d)\n", stream_desc.table_id, stream_desc.no);
			} else {
				dbg(DBG_USR3, "EEPROMRead.readDone: ERROR! Read from eeprom failed (fetching sinlge page)\n");
			}
			ee_state = EE_READY;
		break;
	
		//First page fetched from eeprom, requesting second one
		case EE_READ_FIRST:
			if (success == SUCCESS) {
				if (call EEPROMRead.read(EE_STREAM_START_PAGE + sh.rpos, ee_bufferB) == FAIL) { //TODO mettere controllo sforo limite ee
					dbg(DBG_USR3, "EEPROMRead.readDone: ERROR! Read from eeprom failed (requesting second page)\n");
					ee_state = EE_READY;
				} else {
					ee_state = EE_READ_SECOND;
					dbg(DBG_USR3, "EEPROMRead.readDone: First page fetched from eeprom, requesting second one\n");
				}
			} else {
				dbg(DBG_USR3, "EEPROMRead.readDone: ERROR! Read from eeprom failed (fetching first page)\n");
				ee_state = EE_READY;
			}
		break;

		//Second page fetched, merges with first one and gets the record
		case EE_READ_SECOND:
			if (success == SUCCESS) {
				dbg(DBG_USR3, "EEPROMRead.readDone: Second page fetched from eeprom, merging with first one\n");
				signal SSToApp.readDone(stream_desc, ee_buffer, sh.record_size, SUCCESS);
				sh.rpos += 2;
				sh.stored_records--;
			} else {
				dbg(DBG_USR3, "EEPROMRead.readDone: ERROR! Read from eeprom failed (fetching second page)\n");
			}
			ee_state = EE_READY;
		break;
	}

 /* read_dequeue();
  write_dequeue();*/
  
  return SUCCESS;
}

event result_t EEPROMWrite.writeDone(uint8_t *buffer) {
  call EEPROMWrite.endWrite();
  return SUCCESS;
}

event result_t EEPROMWrite.endWriteDone(result_t success) {

	switch (ee_state) {

      case EE_READY: //debug only
        dbg(DBG_USR3, "EEPROMRead.endWriteDone: ERROR! Inconsistent state! endWriteDone received but no write in progress\n");
      break;

			case EE_WRITE_EMPTY_HEADER:  //when formatting
        if (success == SUCCESS) {
					dbg(DBG_USR3, "EEPROMWrite.endWriteDone: EEprom successfully formatted, eeprom has %d usable pages\n", EE_USABLE_SPACE);
					ee_state = EE_READY;
					ee_write_mode = TRUE;
				} else {
					dbg(DBG_USR3, "EEPROMWrite.endWriteDone: ERROR! eeprom cannot be formatted\n");
          ee_state = EE_NOT_OPENED;
				}
			break;

  		case EE_WRITE_HEADER:   //when closing
  		  if (success == SUCCESS) {
					dbg(DBG_USR3, "EEPROMWrite.endWriteDone: header successfully written\n");
					ee_state = EE_NOT_OPENED;
				} else {
					dbg(DBG_USR3, "EEPROMWrite.endWriteDone: error writing header, eeprom not closed\n");
					ee_state = EE_READY;
				}
			break;

			case EE_WRITE_SINGLE:  //when record size is smaller than eeprom page size
        if (success == SUCCESS) {
          dbg(DBG_USR3, "EEPROMWrite.endWriteDone: page %d successfully written\n", sh.wpos);
          sh.wpos++;
          sh.stored_records++;
				} else {
					dbg(DBG_USR3, "EEPROMWrite.endWriteDone: ERROR! cannot write the page %d\n", sh.wpos);
				}
				ee_state = EE_READY;
			break;

			case EE_WRITE_FIRST: //when record size is bigger than eeprom page size and we are writing the first part
        if (success == SUCCESS) {
          dbg(DBG_USR3, "EEPROMWrite.endWriteDone: first page successfully written ad %d, writing second\n", sh.wpos);
					if (call EEPROMWrite.startWrite() == FAIL) {
            dbg(DBG_USR3, "EEPROMWrite.endWriteDone: ERROR! Cannot write to eeprom (startWrite fails)\n");
            ee_state = EE_READY;
          } else {
						if (call EEPROMWrite.write(EE_STREAM_START_PAGE + sh.wpos, ee_bufferB) == FAIL) {
							dbg(DBG_USR3, "EEPROMWrite.endWriteDone: ERROR! Cannot write to eeprom (write fails)\n");
							ee_state = EE_READY;
						} else {
              ee_state = EE_WRITE_SECOND;
						}
					}
				}
      break;

      case EE_WRITE_SECOND: //when record size is bigger than eeprom page size and we are writing the second part
        if (success == SUCCESS) {
          dbg(DBG_USR3, "EEPROMWrite.endWriteDone: page %d successfully written\n", sh.wpos+1);
          sh.wpos+=2;
          sh.stored_records--;
        }
        ee_state = EE_READY;
      break;
	}

  return SUCCESS;
}


#endif
