#ifndef STREAMBUFFERS_H
#define STREAMBUFFERS_H

#include "BufferPool.h"
#include "SSTypes.h"

#define READ_REQ_SIZE 4

/*
  Protos
*/
inline void sb_init(stream_buffer* sb, uint8_t buffer_size);
inline int8_t sb_write(stream_buffer* sb, int8_t* data, uint8_t nbytes, stream_type stream_desc);
inline result_t sb_read(stream_buffer* sb, stream_type stream_desc);

/*
  Implementation
*/


/*
  Inizializza uno stream buffer

  @param sb Puntatore al descrittore di stream_buffer

  @param buffer_size Numero massimo di elementi memorizzabili nel buffer prima di scartare i più vecchi. 0 = nessun limite
  
*/
inline void sb_init(stream_buffer* sb, uint8_t buffer_size) {
  dbg(DBG_USR2, "sb_init: invoked for stream buffer at %d (buffer_size is %d)\n", sb, buffer_size);
  sb->bp_rpos = NULL;
  sb->bp_wpos = NULL;
  sb->record_size = 0;
  sb->pending_reads = 0;
  sb->buffer_size = buffer_size;
  sb->buffers_used = 0;
}

/*
  Distrugge uno stream buffer eliminando i buffers associati

  @param sb Puntatore al descrittore di stream_buffer
  
*/
inline void sb_destroy(stream_buffer* sb) {
  dbg(DBG_USR2, "sb_destroy: invoked for buffer at %d\n", sb);
  bp_free_all(sb->bp_rpos);
  sb->bp_rpos = NULL;
  sb->bp_wpos = NULL;
  sb->record_size = 0;
  sb->pending_reads = 0;
  sb->buffer_size = 0;
  sb->buffers_used = 0;
}


/*
  Alloca un buffer item dal buffer pool e lo associa allo stream_buffer sb
  quindi copia i dati puntati da data nel nuovo buffer. La dimensione del record
  è stabilita dalla prima sb_write effettuata. Se una successiva sb_write avrà
  nbytes diverso dal primo valore fornito, sarà ritornato un errore e nessun dato sarà
  scritto. Se il valore di nbytes supera BUFFER_SIZE viene ritornato un errore e nessun
  dato sarà scritto.

  @param sb Puntatore al descrittore di stream_buffer

  @param data Puntatore ai dati

  @param nbytes Numero di bytes contenuti nel buffer puntato da data

  @param stream_desc Descrittore di stream da passare a readDone

  @return nbytes in caso di successo, -1 altrimenti

*/
inline int8_t sb_write(stream_buffer* sb, int8_t* data, uint8_t nbytes, stream_type stream_desc) {
  int8_t i;
  bp_item* n;
  
  dbg(DBG_USR2, "sb_write: invoked for stream %d\n", stream_desc.no);
  
  if (nbytes == 0) {
    dbg(DBG_USR2, "sb_write: WARNING: nbytes is zero, ignoring write request\n");
    return 0;
  }
  
  if (nbytes > BUFFER_SIZE) {
    dbg(DBG_USR2, "sb_write: ERROR: nbytes is greater than BUFFER_SIZE (%d > %d)\n", nbytes, BUFFER_SIZE);
    return -1;
  }
  
  if (sb->record_size == 0) {
    dbg(DBG_USR2, "sb_write: record size initialized to %d\n", nbytes);
    sb->record_size = nbytes;
  } else {
    if (nbytes != sb->record_size) {
      dbg(DBG_USR2, "sb_write: ERROR: nbytes differs from record_size (expected %d, found %d)\n", sb->record_size, nbytes);
      return -1;
    }
  }

  n = bp_alloc();
  
  if (n == NULL) {
    dbg(DBG_USR2, "sb_write: ERROR: no space left on buffer pool\n");
    return -1;
  }

  dbg(DBG_USR2, "sb_write: new buffer is at %d (his next should point to NULL: %d)\n", n, n->next);

  for (i=0; i < nbytes; i++)
    n->data[i] = data[i];

  dbg(DBG_USR2, "sb_write: data written at %d: [ %d %d %d %d %d %d %d %d %d %d ... ]\n", n->data, n->data[0],n->data[1],n->data[2],n->data[3],n->data[4],n->data[5],n->data[6],n->data[7],n->data[8],n->data[9]);

	if (sb->bp_wpos == NULL) {
	  dbg(DBG_USR2, "sb_write: bp_wpos is null...\n");
		sb->bp_wpos = n;
	} else {
		sb->bp_wpos->next = n; //Il next dell'ultimo buffer scritto punta al nuovo buffer
		sb->bp_wpos = n;  //L'ultimo buffer scritto diviene il nuovo buffer
	}

  dbg(DBG_USR2, "sb_write: bp_wpos points to %d\n", sb->bp_wpos);

  if (sb->bp_wpos->next!=NULL) {
    dbg(DBG_USR2, "sb_write: ERROR bp_wpos as a next pointing at %d\n", sb->bp_wpos->next);
  }

  sb->buffers_used++;

  if (sb->bp_rpos == NULL) {// bp_rpos può puntare al buffer appena creato (l'unico leggibile)
    sb->bp_rpos = n;
    dbg(DBG_USR2, "sb_write: bp_rpos was null, now points to %d\n", sb->bp_rpos);
  } else {
		//Il numero massimo di buffers utilizzabili per questo stream buffer
		//è stato raggiunto. Si deve scartare il buffer più vecchio, quello
		//puntato da rpos
		if ((sb->buffer_size > 0) && (sb->buffers_used > (sb->buffer_size+1))) { //FIXME è giusto +1? era nel vecchio codice, senza nn va
			n = sb->bp_rpos;
			sb->bp_rpos = sb->bp_rpos->next;
			dbg(DBG_USR2, "sb_write: buffers_used=%d > buffer_size=%d, discarding oldest buffer at %d\n", sb->buffers_used, sb->buffer_size, n);
			bp_free(n);
			sb->buffers_used--;
		}
  }

  if (sb->pending_reads > 0) { // ci sono read pendenti, soddisfa la prima
    dbg(DBG_USR2, "sb_write: there are pending reads, invoking sb_read\n");
    sb_read(sb, stream_desc);
    sb->pending_reads--;
  }

  dbg(DBG_USR2, "sb_write: completed (%d buffers used by this stream)\n", sb->buffers_used);
  return nbytes;

}

/*
  Richiede una lettura dallo stream buffer puntato da sb. Se non vi sono dati disponibili
  la richiesta viene messa in coda e sarà soddisfatta alla successiva write. Il numero di bytes
  ritornato sarà pari alla dimensione di un record (record_size)

  @param sb Puntatore al descrittore di stream_buffer

  @param stream_desc Descrittore di stream da passare a readDone
  
  @return SUCCESS
  
*/
inline result_t sb_read(stream_buffer* sb, stream_type stream_desc) {
  bp_item* t;

  dbg(DBG_USR2, "sb_read: invoked for stream %d\n", stream_desc.no);
  
  if (sb->bp_rpos == NULL) {  // nessun buffer disponibile per la lettura
    sb->pending_reads++;
    dbg(DBG_USR2, "sb_read: no data available, request enqueued (pending_reads=%d), returning\n", sb->pending_reads);
    return SUCCESS;
  }

  dbg(DBG_USR2, "sb_read: signaling readDone (buffer is at %d)\n", sb->bp_rpos->data);
  signal SSToApp.readDone(stream_desc, sb->bp_rpos->data, sb->record_size, SUCCESS);
  dbg(DBG_USR2, "sb_read: readDone reads %d bytes at %d: [ %d %d %d %d %d %d %d %d %d %d ... ]\n", sb->record_size, sb->bp_rpos->data, sb->bp_rpos->data[0],sb->bp_rpos->data[1],sb->bp_rpos->data[2],sb->bp_rpos->data[3],sb->bp_rpos->data[4],sb->bp_rpos->data[5],sb->bp_rpos->data[6],sb->bp_rpos->data[7],sb->bp_rpos->data[8],sb->bp_rpos->data[9]);

  t = sb->bp_rpos;
  
  if (sb->bp_rpos == sb->bp_wpos) {
    dbg(DBG_USR2, "sb_read: no more data on buffer, resetting wpos and rpos to null\n");
    sb->bp_rpos = sb->bp_wpos = NULL;
  } else {
    sb->bp_rpos = sb->bp_rpos->next; // rpos punta al buffer successivo a quello appena letto (nullo se non ve ne sono altri)
    dbg(DBG_USR2, "sb_read: bp_rpos now points to the next buffer at %d (wpos=%d)\n", sb->bp_rpos, sb->bp_wpos);
  }

  bp_free(t);
  sb->buffers_used--;
  dbg(DBG_USR2, "sb_read: completed (%d buffers left for this stream)\n", sb->buffers_used);

  return SUCCESS;

}
  
#endif
